if tonumber(select(4, GetBuildInfo())) < 30100 then
	DisableAddOn"Talented"
	error"Talented v2.0 is only compatible with 3.1 PTR"
	return
else
	InterfaceOptionsDisplayPanelPreviewTalentChanges:Disable()
	SetCVar("previewTalents", "1")
end

local PlayerClass = select(2, UnitClass("player"))

local Talented = LibStub("AceAddon-3.0"):NewAddon("Talented",
	"AceConsole-3.0", "AceComm-3.0", "AceHook-3.0", "AceEvent-3.0", "AceSerializer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Talented")

Talented.max_talent_points = 71

Talented.prev_Print = Talented.Print
function Talented:Print(s, ...)
	if type(s) == "string" and s:find("%", nil, true) then
		self:prev_Print(s:format(...))
	else
		self:prev_Print(s, ...)
	end
end

function Talented:Debug(...)
	if self.db.profile.debug then
		self:Print(...)
	end
end

function Talented:MakeTarget()
	local name = self.db.char.target
	local src = name and self.db.global.templates[name]
	if not src then
		if name then
			self.db.char.target = nil
		end
		return
	end

	local target = self.target
	if not target then
		target = {}
		self.target = target
	end
	self:CopyPackedTemplate(src, target)

	if not self:ValidateTemplate(target) or target.class ~= PlayerClass then
		self.target = nil
		self.db.char.target = nil
	end
	return target
end

function Talented:SetTemplateMode(mode, template)
	if mode then
		if not template then
			local p = self.db.global
			local t = self.db.profile.last_template
			if t then
				template = p.templates[t]
			end
			if not template then
				local _
				_, template = next(p.templates)
			end
		end
		self:SetTemplate(template or self:CreateEmptyTemplate())
	else
		self:SetTemplate()
	end
end

function Talented:OnDisable()
	self:SetTemplateMode()
	self:UnhookInspectUI()
end

function Talented:GetMode()
	return self.mode
end

function Talented:SetMode(mode)
	if self.template == self.current and mode ~= "edit" then
		mode = "view"
	end
	if self.mode ~= mode then
		self.mode = mode
		if mode == "apply" then
			self:ApplyCurrentTemplate()
		elseif mode == "delete" then
			self:DeleteCurrentTemplate()
		elseif self.base and self.base.view then
			self.base.view:SetViewMode(mode)
		end
		self:UpdateView()
	end
	local cb = self.base and self.base.checkbox
	if cb then
		cb:SetChecked(mode == "edit")
	end
end

function Talented:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("TalentedDB", self.defaults)

	self:LoadTemplates()

	local AceDBOptions = LibStub("AceDBOptions-3.0", true)
	if AceDBOptions then
		self.options.args.profiles = AceDBOptions:GetOptionsTable(self.db)
		self.options.args.profiles.order = 200
	end

	LibStub("AceConfig-3.0"):RegisterOptionsTable("Talented", self.options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Talented", "Talented")
	self:RegisterChatCommand("talented", "OnChatCommand")

	self:RegisterComm("Talented")
	if self.InitializePet then
		self:InitializePet()
	end
end

function Talented:OnChatCommand(input)
	if not input or input:trim() == "" then
		self:OpenOptionsFrame()
	else
		LibStub("AceConfigCmd-3.0").HandleCommand(self, "talented", "Talented", input)
	end
end

function Talented:DeleteCurrentTemplate()
	local template = self.template
	if template == self.current then return end
	local templates = self.db.global.templates
	templates[template.name] = nil
	local _
	_, template = next(templates)
	self:SetTemplateMode(template ~= nil, template)
end

function Talented:UpdateTemplateName(template, newname)
	if self.db.global.templates[newname] or
			self.current == template or
			type(newname) ~= "string" or newname == "" then
		return
	end

	local oldname = template.name
	template.name = newname
	local t = self.db.global.templates
	t[newname] = template
	t[oldname] = nil
end

do
	local function new(templates, name, class)
		local count = 0
		local template = {
			name = name,
			class = class
		}
		while templates[template.name] do
			count = count + 1
			template.name = format(L["%s (%d)"], name, count)
		end
		templates[template.name] = template
		return template
	end

	local function copy(dst, src)
		dst.class = src.class
		if src.code then
			dst.code = src.code
			return
		else
			for tab, tree in ipairs(Talented:UncompressSpellData(src.class)) do
				local s, d = src[tab], {}
				dst[tab] = d
				for index = 1, #tree do
					d[index] = s[index]
				end
			end
		end
	end

	function Talented:ImportFromOther(name, src)
		if not self:UncompressSpellData(src.class) then return end

		local dst = new(self.db.global.templates, name, src.class)

		copy(dst, src)

		self:OpenTemplate(dst)

		return dst
	end

	function Talented:CopyTemplate(src)
		local dst = new(self.db.global.templates, format(L["Copy of %s"], src.name), src.class)

		copy(dst, src)

		return dst
	end

	function Talented:CreateEmptyTemplate(class)
		if not class then class = PlayerClass end
		local template = new(self.db.global.templates, L["Empty"], class)

		local info = self:UncompressSpellData(class)

		for tab, tree in ipairs(info) do
			local t = {}
			template[tab] = t
				for index = 1, #tree do
				t[index] = 0
			end
		end

		return template
	end

	Talented.importers = {}
	Talented.exporters = {}
	function Talented:ImportTemplate(url)
		local dst, result = new(self.db.global.templates, L["Imported"])
		for pattern, method in pairs(self.importers) do
			if url:find(pattern) then
				result = method(self, url, dst)
				if result then break end
			end
		end
		if result then
			if not self:ValidateTemplate(dst) then
				self:Print(L["The given template is not a valid one!"])
				self.db.global.templates[dst.name] = nil
			else
				return dst
			end
		else
			self:Print(L["\"%s\" does not appear to be a valid URL!"], url)
		end
	end
end

function Talented:OpenTemplate(template)
	local base = self:CreateBaseFrame()
	if not self.alternates then
		self:UpdateCurrentTemplate()
	end
	self:SetTemplateMode(true, template)
	if not base:IsVisible() then
		ShowUIPanel(base)
	end
end

function Talented:SetTemplate(template)
	if not template then template = self:GetActiveTalents() end
	local view = self:CreateBaseFrame().view
	local old = view.template
	if template ~= old then
		self:UnpackTemplate(template)
		if template.talentGroup and not template.pet then
			view:SetTemplate(template, self:MakeTarget())
		else
			view:SetTemplate(template)
		end
		self:PackTemplate(old)
		self.template = template
	end
	if not template.talentGroup then
		self.db.profile.last_template = template.name
	end
	self:SetMode(self:GetDefaultMode())
	-- self:UpdateView()
end

function Talented:GetDefaultMode()
	return self.db.profile.always_edit and "edit" or "view"
end

function Talented:OnEnable()
	self:RawHook("ToggleTalentFrame", true)
	self:RawHook("ToggleGlyphFrame", true)
	self:SecureHook("UpdateTalentButton")
	self:SecureHook("UpdateMicroButtons")
	self:RegisterEvent("PLAYER_LOGOUT")
	self:CheckHookInspectUI()

	UIParent:UnregisterEvent"GLYPHFRAME_OPEN"
	self:RegisterEvent"GLYPHFRAME_OPEN"
	UIParent:UnregisterEvent"CONFIRM_TALENT_WIPE"
	self:RegisterEvent"CONFIRM_TALENT_WIPE"
	self:RegisterEvent"CHARACTER_POINTS_CHANGED"
	TalentMicroButton:SetScript("OnClick", function () Talented:ToggleTalentFrame() end)
end

function Talented:CONFIRM_TALENT_WIPE(_, cost)
	local dialog = StaticPopup_Show"CONFIRM_TALENT_WIPE"
	if dialog then
		MoneyFrame_Update(dialog:GetName().."MoneyFrame", cost)
		self:SetTemplate()
		local frame = self.base
		if not frame or not frame:IsVisible() then
			self:Update()
			ShowUIPanel(self.base)
		end
		dialog:SetFrameLevel(frame:GetFrameLevel() + 5)
	end
end

function Talented:PLAYER_LOGOUT()
	self:SaveTemplates()
end

function Talented:CHARACTER_POINTS_CHANGED()
	self:UpdateCurrentTemplate()
	self:UpdateView()
	if self.mode == "apply" then
		self:CheckTalentPointsApplied()
	end
end

function Talented:UpdateTalentButton()
	if self.db.profile.donthide and UnitLevel("player") < 10 then
		TalentMicroButton:Show()
		QuestLogMicroButton:SetPoint("BOTTOMLEFT", "TalentMicroButton", "BOTTOMRIGHT", -2, 0)
	end
end

function Talented:UpdateMicroButtons()
	if self.base and self.base:IsShown() then
		TalentMicroButton:SetButtonState("PUSHED", 1)
	else
		TalentMicroButton:SetButtonState"NORMAL"
	end
end

function Talented:ToggleTalentFrame()
	local frame = self.base
	if not frame or not frame:IsVisible() then
		self:Update()
		ShowUIPanel(self.base)
	else
		HideUIPanel(frame)
	end
end

function Talented:Update()
	self:CreateBaseFrame()
	self:UpdateCurrentTemplate()
	if not self.template then
		self:SetTemplate(self.current)
	end
	self:UpdateView()
end

function Talented:LoadTemplates()
	local db = self.db.global.templates
	for name, code in pairs(db) do
		if type(code) == "string" then
			local template = {
				name = name,
				code = code,
			}
			local status = pcall(self.StringToTemplate, self, code, template)
			if not status or not self:ValidateTemplate(template) then
				template = nil
			end
			db[name] = template
		elseif not self:ValidateTemplate(code) then --@FIXME: temporary, while waiting for pet compression
			db[name] = nil
		end
	end
end

function Talented:SaveTemplates()
	local db = self.db.global.templates
	for name, template in pairs(db) do
		template.talentGroup = nil
		self:PackTemplate(template)
		if template.code then
			db[name] = template.code
		end
	end
end

_G.Talented = Talented
